/* 
   $Id: sgbar.c,v 1.3 2001/12/25 05:15:34 sho Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include "sgwmwidget.h"
#include "sgbar.h"
#include <gtk/gtkwindow.h>
#include "sgbarwindow.h"
#include "sgbarholder.h"

static void sgbar_class_init(SGBarClass *klass);
static void sgbar_init(SGBar *win);
static GtkWidget *parent_class;

GtkType sgbar_get_type(void)
{
    static GtkType sgbar_type = 0;

    if (sgbar_type == 0) {
	GtkTypeInfo sgbar_info = {
	    "SGBar",
	    sizeof(SGBar),
	    sizeof(SGBarClass),
	    (GtkClassInitFunc)sgbar_class_init,
	    (GtkObjectInitFunc)sgbar_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgbar_type = gtk_type_unique(gtk_bin_get_type(),
	    &sgbar_info);
    }

    return sgbar_type;
}

static void sgbar_size_request(GtkWidget *widget, GtkRequisition *req)
{
    GtkBin *bin = GTK_BIN(widget);

    req->width = 0;
    req->height = 0;
    if (bin->child != NULL) {
	gtk_widget_size_request(bin->child,req);
    }
}

static void sgbar_size_allocate(GtkWidget *widget, GtkAllocation *alloc)
{
    GtkBin *bin = GTK_BIN(widget);
    gint bw;
    GtkAllocation a;

/*
    if (widget->allocation.x == alloc->x
      && widget->allocation.y == alloc->y
      && widget->allocation.width == alloc->width
      && widget->allocation.height == alloc->height) return;
      */
    widget->allocation = *alloc;
    if (GTK_WIDGET_REALIZED(widget)) {
	gdk_window_move_resize(widget->window,alloc->x,alloc->y,
	  alloc->width,alloc->height);
    }
    if (bin->child != NULL) {
	SGBar *bar = SGBAR(widget);

	bw = GTK_CONTAINER(widget)->border_width;
	a.x = bw;
	a.y = bw;
	a.width = alloc->width - bw*2;
	a.height = alloc->height - bw*2;
	if (bar->movable) {
	    a.x += SG_BAR_HANDLE_WIDTH;
	}
	gtk_widget_size_allocate(bin->child,&a);
    }
}

static void sgbar_draw(GtkWidget *widget, GdkRectangle *area)
{
    GtkBin *bin = GTK_BIN(widget);
    GdkRectangle child_area;

    if (bin->child != NULL) {
	SGBar *bar = SGBAR(widget);

	if (bar->movable && area->x >= 0
	  && area->x < SG_BAR_HANDLE_WIDTH) {
	    gtk_paint_handle(widget->style, widget->window, GTK_STATE_NORMAL,
	      GTK_SHADOW_OUT, area, widget, "sgbar",
	      0, 0, SG_BAR_HANDLE_WIDTH, widget->allocation.height, 
	      GTK_ORIENTATION_VERTICAL);

	}
	if (gtk_widget_intersect(bin->child,area,&child_area)) {
	    gtk_widget_draw(bin->child,&child_area);
	}
    }
}

static gint sgbar_expose(GtkWidget *widget, GdkEventExpose *event)
{
    GtkBin *bin = GTK_BIN(widget);
    GdkRectangle area;

    if (bin->child != NULL) {
	SGBar *bar = SGBAR(widget);

	if (bar->movable && event->area.x >= 0
	  && event->area.x < SG_BAR_HANDLE_WIDTH) {
	    gtk_paint_handle(widget->style, widget->window, GTK_STATE_NORMAL,
	      GTK_SHADOW_OUT, &(event->area), widget, "sgbar",
	      0, 0, SG_BAR_HANDLE_WIDTH, widget->allocation.height, 
	      GTK_ORIENTATION_VERTICAL);

	}
	if (gtk_widget_intersect(bin->child,&(event->area),&area)) {
	    gtk_widget_draw(bin->child,&area);
	}
    }
    return TRUE;
}

void sgbar_set_movable(SGBar *bar, gboolean movable)
{
    GdkEventMask mask;
    GtkWidget *widget;

    if (bar->movable == movable) return;
    bar->movable = movable;
    widget = GTK_WIDGET(bar);
    mask = gdk_window_get_events(widget->window);
    if (movable) {
	mask |= GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	   | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK;
    } else {
	mask &= ~(GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	   | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
    }
    gdk_window_set_events(widget->window,mask);
}

static void sgbar_realize(GtkWidget *widget)
{
    GdkWindowAttr attributes;
    gint attributes_mask;
    gint bw;
    SGBar *bar;

    bar = SGBAR(widget);
    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);

    bw = GTK_CONTAINER(widget)->border_width;
    attributes.x = widget->allocation.x+bw;
    attributes.y = widget->allocation.y+bw;
    attributes.width = widget->allocation.width-bw*2;
    attributes.height = widget->allocation.height-bw*2;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);
    attributes.event_mask = GDK_EXPOSURE_MASK;
    if (bar->movable) {
	attributes.event_mask |= GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	   | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK;
    }
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new(gtk_widget_get_parent_window(widget),
      &attributes,attributes_mask);
    gdk_window_set_user_data(widget->window,widget);
    widget->style = gtk_style_attach(widget->style, widget->window);
    gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);
}

static gint sgbar_button_press(GtkWidget *widget, GdkEventButton *event)
{
    SGBar *bar = SGBAR(widget);
    GtkBin *bin = GTK_BIN(widget);

    if (bar->movable && !bar->in_drag && bin->child != NULL && event->button == 1
       && event->x >= 0 && event->x < SG_BAR_HANDLE_WIDTH) {
	bar->in_drag = TRUE;
	bar->oldx = widget->allocation.x;
    }
    return TRUE;
}

static gint sgbar_button_release(GtkWidget *widget, GdkEventButton *event)
{
    SGBar *bar = SGBAR(widget);
    GtkBin *bin = GTK_BIN(widget);

    if (bar->in_drag && event->button == 1) {
	bar->in_drag = FALSE;
	if (bin->child != NULL) {
	    SGBarWindow *bwin = SGBARWINDOW(bin->child);
	    gint x;

	    sgbarholder_get_x(widget->parent,&x);
	    widget->allocation.x = x;
	    gdk_window_move(widget->window,widget->allocation.x,
	      widget->allocation.y);
	    bwin->x = x;
	    bwin->moved = TRUE;
	}
    }
    return TRUE;
}

static gint sgbar_motion(GtkWidget *widget, GdkEventMotion *event)
{
    SGBar *bar = SGBAR(widget);
    gint x;

    if (!bar->in_drag) return TRUE;

    sgbarholder_get_x(widget->parent,&x);
    widget->allocation.x = x;
    gdk_window_move(widget->window,widget->allocation.x,
      widget->allocation.y);
    return TRUE;
}

static void sgbar_class_init(SGBarClass *klass)
{
    GtkWidgetClass *widget_class;

    parent_class = gtk_type_class(gtk_bin_get_type());
    widget_class = (GtkWidgetClass *)klass;
    widget_class->realize = sgbar_realize;
    widget_class->draw = sgbar_draw;
    widget_class->expose_event = sgbar_expose;
    widget_class->button_press_event = sgbar_button_press;
    widget_class->button_release_event = sgbar_button_release;
    widget_class->motion_notify_event = sgbar_motion;
    widget_class->size_allocate = sgbar_size_allocate;
    widget_class->size_request = sgbar_size_request;
}

static void sgbar_init(SGBar *bar)
{
    GTK_WIDGET_UNSET_FLAGS(bar,GTK_NO_WINDOW);
    gtk_container_set_resize_mode(GTK_CONTAINER(bar),GTK_RESIZE_QUEUE);
    bar->in_drag = FALSE;
    bar->movable = FALSE;
}

GtkWidget *sgbar_new()
{
    GtkWidget *widget;

    widget  = GTK_WIDGET(gtk_type_new(sgbar_get_type()));
    return widget;
}

gboolean sgbar_get_x(SGBar *bar, gint *x)
{
    GtkBin *bin = GTK_BIN(bar);

    if (bin->child != NULL) {
	SGBarWindow *bwin = SGBARWINDOW(bin->child);

	if (bwin->moved) {
	    *x = bwin->x;
	    return TRUE;
	}
    }
    return FALSE;
}
