/* 
   $Id: sgwmwidget.c,v 1.14 2001/12/27 06:15:00 sho Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkbindings.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkwindow.h>
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include "sgwmwidget.h"
#include "sgwindow.h"
#include "sgwmtop.h"
#include "atoms.h"
#include "sgintl.h"

typedef GdkFilterReturn (*Handler)(SGWMWidget *sgwm, XEvent *ev);

static Handler handlers[LASTEvent];

static void sgwmwidget_class_init(SGWMWidgetClass *klass);
static void sgwmwidget_init(SGWMWidget *sgwm);
static Cursor queryWSCursor;
static Cursor defaultCursor;

#ifndef MIDDLE_BUTTON_KEYCODE
#define MIDDLE_BUTTON_KEYCODE 0
#endif
#ifndef RIGHT_BUTTON_KEYCODE
#define RIGHT_BUTTON_KEYCODE 0
#endif

gint middleButtonKeycode = MIDDLE_BUTTON_KEYCODE;
gint rightButtonKeycode = RIGHT_BUTTON_KEYCODE;

#define SG_MAX_N_BUTTONS 10

static gint nbuttons = 0;

#define SG_BUTTON_LEFT 0
#define SG_BUTTON_MIDDLE 1
#define SG_BUTTON_RIGHT 2

static gint buttonState = SG_BUTTON_LEFT;

enum {
    SWAP_TASK,
    ENABLE_WORKSPACE,
    ACTIVATE_WORKSPACE,
    RESIZE_WORKSPACE,
    CALL_DESKTOP,
    MAP_TASK_TO_WORKSPACE,
    MAP_TASK,
    ROLL_UP_TASK,
    ROLL_DOWN_TASK,
    MOVE_NEXT_TASK,
    MOVE_PREV_TASK,
    LAST_SIGNAL
};

static guint sgwm_signals[LAST_SIGNAL] = {0};

GtkType sgwmwidget_get_type(void)
{
    static GtkType sgwmwidget_type = 0;

    if (sgwmwidget_type == 0) {
	GtkTypeInfo sgwmwidget_info = {
	    "SGWMWidget",
	    sizeof(SGWMWidget),
	    sizeof(SGWMWidgetClass),
	    (GtkClassInitFunc)sgwmwidget_class_init,
	    (GtkObjectInitFunc)sgwmwidget_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgwmwidget_type = gtk_type_unique(gtk_widget_get_type(),
	    &sgwmwidget_info);
    }

    return sgwmwidget_type;
}

static void sgwmwidget_call_desktop(SGWMWidget *sgwm, gint no)
{
    SGWMTop *top = SGWMTOP(sgwm->top);

    if (sgwm->desktop == NULL) return;
    sgwmtop_call_desktop(top,SGWINDOW(sgwm->desktop),no);
}

static void sgwmwidget_set_client_list_property(SGWMWidget *sgwm)
{
    Window *list = NULL;
    int n;
    int i;
    GList *cur;

    if (!sgwm->client_list_changed) return;
    n = g_list_length(sgwm->client_list);
    if (n <= 0) {
	XChangeProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),xa_win_client_list,
	  XA_CARDINAL,32,PropModeReplace,NULL,0);
    } else if ((list = g_malloc(g_list_length(sgwm->client_list)
      *sizeof(Window))) != NULL) {
	for (i = 0,cur = sgwm->client_list;i < n;i++, cur = cur->next) {
	    list[i] = (Window)cur->data;
	}
	XChangeProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),xa_win_client_list,
	  XA_CARDINAL,32,PropModeReplace,(unsigned char *)list,n);
	g_free(list);
    }
    sgwm->client_list_changed = FALSE;
}

static gboolean sgwmwidget_is_desktop(Window xwin)
{
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    gboolean r = FALSE;

    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_desktop,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	XFree(rv);
	r = TRUE;
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    return r;
}

static gboolean sgwmwidget_is_barwindow(Window xwin)
{
    gint rval = -1;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;

    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_barwindow,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	if (rf == 32 && rni == 1) {
	    rval = ((gint *)rv)[0];
	}
	XFree(rv);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    return rval != -1;
}

static gint sgwmwidget_get_panel_property(Window xwin)
{
    gint rval = -1;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;

    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_panel,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	if (rf == 32 && rni == 1) {
	    rval = ((gint *)rv)[0];
	}
	XFree(rv);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    return rval;
}

static void sgwmwidget_map_nunome(SGWMWidget *sgwm, gint xid)
{
    GtkWidget *widget;

    widget = sgwindow_lookup(xid,sgwm);
    if (widget != NULL) {
	SGWindow *sgwin = SGWINDOW(widget);

	sgwin->type = SGWINDOW_NUNOME;
	sgwmtop_map_nunome(SGWMTOP(sgwm->top),sgwin);
    }
}

static gboolean sgwmwidget_is_nunome(Window win)
{
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    gboolean r = FALSE;

    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),win,xa_sg_nunome,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	r = TRUE;
	XFree(rv);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	r = FALSE;
    }
    return r;
}

static void sgwmwidget_resize_top(SGWMWidget *sgwm, gint px, gint py,
  gint pw, gint ph)
{
    GtkWidget *widget;
    GtkAllocation alloc;

    widget = GTK_WIDGET(sgwm);
    if (px >= 0 && py >= 0) {
	if (pw == widget->allocation.width
          || ph == widget->allocation.height) {
	    if (pw == widget->allocation.width) {
		if (py == 0) {
		    alloc.x = 0;
		    alloc.y = py+ph;
		    alloc.width = widget->allocation.width;
		    alloc.height = widget->allocation.height-alloc.y;
		} else {
		    alloc.x = 0;
		    alloc.y = 0;
		    alloc.width = widget->allocation.width;
		    alloc.height = py;
		}
	    } else {
		if (px == 0) {
		    alloc.x = px+pw;
		    alloc.y = 0;
		    alloc.width = widget->allocation.width-alloc.x;
		    alloc.height = widget->allocation.height;
		} else {
		    alloc.x = 0;
		    alloc.y = 0;
		    alloc.width = px;
		    alloc.height = widget->allocation.height;
		}
	    }
	} else {
	    alloc.x = 0;
	    alloc.y = 0;
	    alloc.width = widget->allocation.width;
	    alloc.height = widget->allocation.height;
	}
#ifdef __SIKIGAMI_PDA__
	gtk_widget_set_uposition(sgwm->top, alloc.x, alloc.y);
#endif
	gtk_widget_size_allocate(sgwm->top,&alloc);
    }
}

static void sgwmwidget_map_panel(SGWMWidget *sgwm, Window win, int panel_type)
{
    gdk_error_trap_push();
    if (panel_type == 0) {
	Window rw;
	int x,y;
	guint w,h,bw,d;
	long supplied;
	XSizeHints *hints;

	XGetGeometry(GDK_DISPLAY(),win,&rw,&x,&y,&w,&h,&bw,&d);
	if ((hints = XAllocSizeHints()) != NULL) {
	    if (XGetWMNormalHints(GDK_DISPLAY(),win,hints,&supplied)) {
		hints->flags &= supplied;
		if (hints->flags & (USPosition | PPosition)) {
		    x = hints->x;
		    y = hints->y;
		}
		XMoveWindow(GDK_DISPLAY(),win,x,y);
	    }
	    XFree(hints);
	}
	sgwmwidget_resize_top(sgwm,x,y,w,h);
	gdk_flush();
    }
    XMapWindow(GDK_DISPLAY(),win);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
}

static int sgwmwidget_get_workspace(SGWMWidget *sgwm, SGWindow *win,
  gboolean query)
{
    int pstat,kstat;
    XEvent event;
    int no;
    SGWMTop *top = SGWMTOP(sgwm->top);

    if ((no = sgwindow_get_workspace_property(win)) >= 0) {
	return no;
    }
    if (sgwindow_get_transient_for(win) != 0) {
	if ((no = sgwmtop_get_workspace_transient(top,win)) >= 0) {
	    return no;
	}
    }
    if (sgwmtop_get_enabled_workspace_count(SGWMTOP(sgwm->top),&no) == 1) {
	return no;
    }
    if (!query) return 0;
    kstat = pstat = AlreadyGrabbed;
    do {
	XUngrabServer(GDK_DISPLAY());
	if (pstat == GrabSuccess) {
	    XUngrabPointer(GDK_DISPLAY(),CurrentTime);
	}
	if (kstat == GrabSuccess) {
	    XUngrabKeyboard(GDK_DISPLAY(),CurrentTime);
	}
	gdk_flush();
	XGrabServer(GDK_DISPLAY());
	pstat = XGrabPointer(GDK_DISPLAY(),GDK_ROOT_WINDOW(),False,
		ButtonPressMask,
		GrabModeAsync, GrabModeAsync,
		GDK_ROOT_WINDOW(),queryWSCursor, CurrentTime);
	kstat = XGrabKeyboard(GDK_DISPLAY(),GDK_ROOT_WINDOW(),False,
		GrabModeAsync, GrabModeAsync,CurrentTime);
    } while (pstat != GrabSuccess || kstat != GrabSuccess);
    gtk_widget_show_all(sgwm->prompt);
    gtk_widget_draw(sgwm->prompt,NULL);
    do {
	XMaskEvent(GDK_DISPLAY(), ButtonPressMask | KeyPressMask , &event);
	if (event.type == ButtonPress) {
	    no = sgwmtop_get_workspace_coord(top,event.xbutton.x_root,
	      event.xbutton.y_root);
	} else if (event.type == KeyPress) {
	    unsigned int keysym;

	    keysym = XLookupKeysym(&(event.xkey),0);
	    switch (keysym) {
	    case XK_0:
		no = 0;
		break;
	    case XK_1:
		no = 1;
		break;
	    case XK_2:
	    case XK_3:
	    case XK_4:
	    case XK_5:
	    case XK_6:
	    case XK_7:
	    case XK_8:
	    case XK_9:
		break;
	    }
	}
    } while (no < 0);
    gtk_widget_hide(sgwm->prompt);
    XUngrabPointer(GDK_DISPLAY(),CurrentTime);
    XUngrabKeyboard(GDK_DISPLAY(),CurrentTime);
    XUngrabServer(GDK_DISPLAY());
    return no;
}

static GtkWidget *sgwmwidget_client_lookup(SGWMWidget *sgwm, gint xid)
{
    GdkWindow *win;

    if ((win = gdk_window_lookup(xid)) == NULL) {
	sgwm->client_list = g_list_prepend(sgwm->client_list,
	  (gpointer)xid);
	sgwm->client_list_changed = TRUE;
	gdk_error_trap_push();
	XAddToSaveSet(GDK_DISPLAY(),xid);
	gdk_flush();
	if (gdk_error_trap_pop()) {
	    ;
	}
    }
    return sgwindow_from_gdk_window(win,xid,sgwm);
}

static gboolean sgwmwidget_map_task_iconic(SGWMWidget *sgwm,
  SGWindow *sgwin)
{
    if (sgwin->workspace < 0) {
	XWMHints *p;

	p = sgwindow_get_wm_hints(sgwin);
	if (p != NULL && (p->flags & StateHint) != 0
	   && p->initial_state == IconicState) {
	    sgwmtop_map_task_iconic(SGWMTOP(sgwm->top),sgwin);
	    return TRUE;
	}
    }
    return FALSE;
}

static void sgwmwidget_map_task_to_workspace(SGWMWidget *sgwm, gint xid, gint ws)
{
    GtkWidget *widget;
    SGWindow *sgwin;

    widget = sgwmwidget_client_lookup(sgwm,xid);
    if (widget != NULL) {
	sgwin = SGWINDOW(widget);
	if (sgwmwidget_map_task_iconic(sgwm,sgwin)) {
	    goto exiting;
	}
	sgwmtop_map_task(SGWMTOP(sgwm->top),sgwin,ws);
    }
exiting:
    sgwmwidget_set_client_list_property(sgwm);
}

static void sgwmwidget_do_unmap_task(SGWMWidget *sgwm, gint xid)
{
    SGWindow *sgwin;
    GtkWidget *widget;

    widget = sgwmwidget_client_lookup(sgwm,xid);
    if (widget != NULL) {
	sgwin = SGWINDOW(widget);
	sgwmtop_unmap_task(SGWMTOP(sgwm->top),sgwin);
    }
}

static void sgwmwidget_map_desktop(SGWMWidget *sgwm, gint xid)
{
    GtkWidget *widget;

    widget = sgwindow_lookup(xid,sgwm);
    if (widget != NULL) {
	int no;
	SGWindow *sgwin;

	sgwm->desktop = widget;
	sgwin = SGWINDOW(widget);
	if (sgwmwidget_map_task_iconic(sgwm,sgwin)) {
	    return;
	}
	no = sgwmwidget_get_workspace(sgwm,sgwin,FALSE);
	sgwmtop_map_task(SGWMTOP(sgwm->top),sgwin,no);
    }
}

static void sgwmwidget_do_map_task(SGWMWidget *sgwm, gint xid)
{
    SGWindow *sgwin;
    int panel_type;
    GtkWidget *widget;

    if ((panel_type = sgwmwidget_get_panel_property(xid)) >= 0) {
	sgwmwidget_map_panel(sgwm,xid,panel_type);
	return;
    }
    if (sgwmwidget_is_desktop(xid)) {
	sgwmwidget_map_desktop(sgwm,xid);
	return;
    }
    if (sgwmwidget_is_nunome(xid)) {
	sgwmwidget_map_nunome(sgwm,xid);
	return;
    }
    widget = sgwmwidget_client_lookup(sgwm,xid);
    if (widget != NULL) {
	int no;

	sgwin = SGWINDOW(widget);
	if (sgwmwidget_map_task_iconic(sgwm,sgwin)) {
	    return;
	}
	no = sgwmwidget_get_workspace(sgwm,sgwin,TRUE);
	sgwmtop_map_task(SGWMTOP(sgwm->top),sgwin,no);
	if (widget->window == 0) {
	    sgwmwidget_destroy_sgwin(sgwm,widget,xid);
	}
    }
    sgwmwidget_set_client_list_property(sgwm);
}

static void sgwmwidget_map_task(SGWMWidget *sgwm, gint xid)
{
    sgwmwidget_do_map_task(sgwm,xid);
}

static void sgwmwidget_enable_workspace(SGWMWidget *sgwm, gint ws, gint enable)
{
    sgwmtop_enable_workspace(SGWMTOP(sgwm->top),ws,enable);
}

static void sgwmwidget_swap_task(SGWMWidget *sgwm, gint ws1, gint ws2)
{
    sgwmtop_swap_task(SGWMTOP(sgwm->top),ws1,ws2);
}

static void sgwmwidget_resize_workspace(SGWMWidget *sgwm, gint ws,
  gint mode, gint size)
{
    sgwmtop_resize_workspace(SGWMTOP(sgwm->top),ws,mode,size);
}

static void sgwmwidget_roll_up_task(SGWMWidget *sgwm, gint ws)
{
    sgwmtop_roll_up_task(SGWMTOP(sgwm->top),ws);
}

static void sgwmwidget_roll_down_task(SGWMWidget *sgwm, gint ws)
{
    sgwmtop_roll_down_task(SGWMTOP(sgwm->top),ws);
}

static void sgwmwidget_move_next_task(SGWMWidget *sgwm, gint ws)
{
    sgwmtop_move_next_task(SGWMTOP(sgwm->top),ws);
}

static void sgwmwidget_move_prev_task(SGWMWidget *sgwm, gint ws)
{
    sgwmtop_move_prev_task(SGWMTOP(sgwm->top),ws);
}

GdkFilterReturn sgwmwidget_filter(GdkXEvent *xevent,
  GdkEvent *event, gpointer data)
{
    GdkFilterReturn r;
    SGWMWidget *sgwm = SGWMWIDGET(data);
    XEvent *ev = (XEvent *)xevent;

    r = GDK_FILTER_CONTINUE;
    if (ev->type >= 0 || ev->type < LASTEvent) {
	if (ev->xany.send_event) {
#ifdef DEBUG_PRINT
	    printf("sent event\n");
#endif
	}
	if (handlers[ev->type] != 0) {
	    r = (*(handlers[ev->type]))(sgwm,ev);
	}
    }
    return r;
}


static GdkFilterReturn sgwm_handle_key_release(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("key_release\n");
#endif
    if (ev->xkey.keycode == rightButtonKeycode
       || ev->xkey.keycode == middleButtonKeycode) {
	if (buttonState != SG_BUTTON_LEFT) {
	    int i;
	    unsigned char map[SG_MAX_N_BUTTONS];

	    buttonState = SG_BUTTON_LEFT;
	    for (i =0;i < nbuttons;i++) {
		map[i] = i + 1;
	    }
	    XSetPointerMapping(GDK_DISPLAY(),map,nbuttons);
	}
    }
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_key_press(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("key_press\n");
#endif
    if (ev->xkey.keycode == rightButtonKeycode) {
	if (buttonState != SG_BUTTON_RIGHT) {
	    int i;
	    unsigned char map[SG_MAX_N_BUTTONS];

	    buttonState = SG_BUTTON_RIGHT;
	    for (i =0;i < nbuttons;i++) {
		map[i] = i + 3;
	    }
	    XSetPointerMapping(GDK_DISPLAY(),map,nbuttons);
	}
    } else if (ev->xkey.keycode == middleButtonKeycode) {
	if (buttonState != SG_BUTTON_MIDDLE) {
	    int i;
	    unsigned char map[SG_MAX_N_BUTTONS];

	    buttonState = SG_BUTTON_MIDDLE;
	    for (i =0;i < nbuttons;i++) {
		map[i] = i + 2;
	    }
	    XSetPointerMapping(GDK_DISPLAY(),map,nbuttons);
	}
    }
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_button_press(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("button_press\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_button_release(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("button_release\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_motion_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("motion_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_enter_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("enter_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_leave_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("leave_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_expose(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("expose\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_visibility_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("visibility_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_create_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("create_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_destroy_notify(SGWMWidget *sgwm,XEvent *ev)
{
    GdkWindow *win;

#ifdef DEBUG_PRINT
    printf("destroy_notify\n");
#endif
    sgwm->client_list = g_list_remove(sgwm->client_list,
      (gpointer)ev->xdestroywindow.window);
    sgwm->client_list_changed = TRUE;
    sgwmwidget_set_client_list_property(sgwm);
    if (ev->xdestroywindow.event == GDK_ROOT_WINDOW()) {
	return GDK_FILTER_REMOVE;
    }
    win = gdk_window_lookup(ev->xdestroywindow.window);
    if (win != NULL) {
	gpointer p;

	gdk_window_get_user_data(win,&p);
	if (IS_SGWINDOW(p)) {
	    SGWindow *sgwin = (SGWindow *)p;
	    SGWMTop *top = SGWMTOP(sgwm->top);

	    GTK_WIDGET_UNSET_FLAGS(sgwin, GTK_MAPPED | GTK_REALIZED | GTK_VISIBLE);
	    gtk_widget_ref(GTK_WIDGET(p));
	    gdk_error_trap_push();
	    sgwmtop_unmap_task(top,sgwin);
	    sgwmtop_destroy_child_notify(top,sgwin);
	    sgwindow_destroy_notify(sgwin);
	    gtk_widget_unref(GTK_WIDGET(p));
	    gdk_flush();
	    if (gdk_error_trap_pop()) {
		;
	    }
	}
    }
    return GDK_FILTER_REMOVE;
}

void sgwmwidget_destroy_sgwin(SGWMWidget *sgwm, GtkWidget *widget, guint xid)
{
    SGWindow *sgwin = SGWINDOW(widget);
    SGWMTop *top = SGWMTOP(sgwm->top);
    gint x,y;

    if (sgwm->desktop == widget) {
	sgwm->desktop = NULL;
    }
    GTK_WIDGET_UNSET_FLAGS(widget, GTK_MAPPED | GTK_REALIZED | GTK_VISIBLE);
    gtk_widget_ref(widget);
    gdk_error_trap_push();
    sgwmtop_unmap_task(top,sgwin);
    sgwm->client_list = g_list_remove(sgwm->client_list,
      (gpointer)xid);
    sgwm->client_list_changed = TRUE;
    sgwmwidget_set_client_list_property(sgwm);
    sgwindow_change_wm_state(sgwin,WithdrawnState);
    sgwindow_get_position(sgwin,&x,&y);
    sgwmtop_destroy_child_notify(top,sgwin);
    sgwindow_destroy_notify(sgwin);
    XRemoveFromSaveSet(GDK_DISPLAY(),xid);
    gtk_widget_unref(widget);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
}

static GdkFilterReturn sgwm_handle_unmap_notify(SGWMWidget *sgwm,XEvent *ev)
{
    GdkWindow *win;

#ifdef DEBUG_PRINT
    printf("unmap_notify\n");
#endif
    if (ev->xunmap.event == GDK_ROOT_WINDOW()) {
	return GDK_FILTER_REMOVE;
    }
    win = gdk_window_lookup(ev->xunmap.window);
    if (win != NULL) {
	gpointer p;

	gdk_window_get_user_data(win,&p);
	if (IS_SGWINDOW(p)) {
	    sgwmwidget_destroy_sgwin(sgwm,(GtkWidget *)p,ev->xunmap.window);
	}
    }
    return GDK_FILTER_REMOVE;
}

static GdkFilterReturn sgwm_handle_map_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("map_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_map_request(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("map_request\n");
#endif
    sgwmwidget_do_map_task(sgwm,ev->xmaprequest.window);
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwmwidget_configure_panel(SGWMWidget *sgwm,
  XConfigureRequestEvent *cev, gint panel_type)
{
    XWindowChanges xwc;

    xwc.x = cev->x;
    xwc.y = cev->y;
    xwc.width = cev->width;
    xwc.height = cev->height;
    xwc.sibling = cev->above;
    xwc.stack_mode = cev->detail;
    XConfigureWindow(GDK_DISPLAY(),cev->window,cev->value_mask,
      &xwc);
    if (panel_type == 0) {
	sgwmwidget_resize_top(sgwm,cev->x,cev->y,cev->width,cev->height);
    }
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_configure_request(SGWMWidget *sgwm,XEvent *ev)
{
    XConfigureRequestEvent *cev;
    gint panel_type;
    GdkFilterReturn r;

#ifdef DEBUG_PRINT
    printf("configure_request\n");
#endif
    cev = &(ev->xconfigurerequest);
    gdk_error_trap_push();
    if ((panel_type =
      sgwmwidget_get_panel_property(cev->window)) >= 0) {
	r = sgwmwidget_configure_panel(sgwm,cev,panel_type);
    } else {
	XWindowChanges xwc;
	SGWindow *sgwin;
	GtkAllocation newalloc;
	GdkWindow *win;
	GtkWidget *tw;

	newalloc.x = cev->x;
	newalloc.y = cev->y;
	newalloc.width = cev->width;
	newalloc.height = cev->height;
	if ((win = gdk_window_lookup(cev->window)) != NULL
	   && (tw = sgwindow_from_gdk_window(win,cev->window,sgwm)) != NULL) {
	    sgwin = SGWINDOW(tw);
	    sgwmtop_calculate_win_size(SGWMTOP(sgwm->top),sgwin,&newalloc);
	    sgwindow_configure(sgwin,&newalloc,cev->border_width,
	      cev->above, cev->detail, cev->value_mask);
	} else {
	    xwc.x = newalloc.x;
	    xwc.y = newalloc.y;
	    xwc.width = newalloc.width;
	    xwc.height = newalloc.height;
	    xwc.sibling = cev->above;
	    xwc.stack_mode = cev->detail;
	    xwc.border_width = cev->border_width;
	    XConfigureWindow(GDK_DISPLAY(),cev->window,cev->value_mask,
	      &xwc);
	}
	r = GDK_FILTER_CONTINUE;
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    return r;
}

static GdkFilterReturn sgwm_handle_property_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("property_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_colormap_notify(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("colormap_notify\n");
#endif
    return GDK_FILTER_CONTINUE;
}

static void sgwmwidget_activate_workspace(SGWMWidget *sgwm, gint no)
{
    sgwmtop_activate_workspace(SGWMTOP(sgwm->top),no,TRUE,TRUE);
}

static GdkFilterReturn sgwm_handle_client_message(SGWMWidget *sgwm,XEvent *ev)
{
    XClientMessageEvent *cev = (XClientMessageEvent *)ev;

#ifdef DEBUG_PRINT
    printf("client_message\n");
#endif
    if (cev->message_type == xa_sgws_swap) {
	sgwmwidget_swap_task(sgwm,cev->data.l[0],cev->data.l[1]);
    } else if (cev->message_type == xa_sgws_roll_up) {
	sgwmwidget_roll_up_task(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_roll_down) {
	sgwmwidget_roll_down_task(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_move_next) {
	sgwmwidget_move_next_task(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_move_prev) {
	sgwmwidget_move_prev_task(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_activate) {
	sgwmwidget_activate_workspace(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_enable) {
	sgwmwidget_enable_workspace(sgwm,cev->data.l[0],cev->data.l[1]);
    } else if (cev->message_type == xa_sgws_map) {
	sgwmwidget_map_task(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_calldt) {
	sgwmwidget_call_desktop(sgwm,cev->data.l[0]);
    } else if (cev->message_type == xa_sgws_mapws) {
	sgwmwidget_map_task_to_workspace(sgwm,cev->data.l[0],cev->data.l[1]);
    } else if (cev->message_type == xa_win_workspace) {
	if (cev->window == GDK_ROOT_WINDOW()) {
	    sgwmwidget_activate_workspace(sgwm,cev->data.l[0]);
	} else {
	    sgwmwidget_map_task_to_workspace(sgwm,cev->window,cev->data.l[0]);
	}
    } else if (cev->message_type == xa_wm_change_state) {
	if (cev->data.l[0] == IconicState) {
	    sgwmwidget_do_unmap_task(sgwm,cev->window);
	}
    }
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_client_message_filter(GdkXEvent *ev, GdkEvent *gev,
  gpointer data)
{
    SGWMWidget *sgwm = (SGWMWidget *)data;

    return sgwm_handle_client_message(sgwm,(XEvent *)ev);
}

static GdkFilterReturn sgwm_handle_selection_clear(SGWMWidget *sgwm,XEvent *ev)
{
#ifdef DEBUG_PRINT
    printf("selection_clear\n");
#endif
    if (ev->xselectionclear.selection == xa_wm_s) {
	g_error("I lost WM_S0 selection, now I'm exiting....\n");
	sgwmwidget_exit(1);
    }
    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn sgwm_handle_selection_request(SGWMWidget *sgwm,XEvent *ev)
{
    XSelectionEvent sev;
    Atom reply = None;

#ifdef DEBUG_PRINT
    printf("selection_request\n");
#endif
    gdk_error_trap_push();
    if (ev->xselectionrequest.selection == xa_wm_s) {
	if (ev->xselectionrequest.target == xa_version) {
	    unsigned int data[2];

	    reply = ev->xselectionrequest.property;
	    if (reply == None) {
		reply = xa_version;
	    }
	    data[0] = ICCCM_VERSION_MAJOR;
	    data[1] = ICCCM_VERSION_MINOR;
	    XChangeProperty(GDK_DISPLAY(),ev->xselectionrequest.requestor,
	      reply,XA_INTEGER,32,PropModeReplace,(unsigned char *)data,2);
	} 
    }
    sev.type = SelectionNotify;
    sev.requestor = ev->xselectionrequest.requestor;
    sev.selection = xa_wm_s;
    sev.target = xa_version;
    sev.property = reply;
    sev.time = ev->xselectionrequest.time;
    XSendEvent(GDK_DISPLAY(),ev->xselectionrequest.requestor,
      False,0, (XEvent *)&sev);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    return GDK_FILTER_CONTINUE;
}

static void sgwmwidget_init_handlers()
{
    handlers[KeyPress] = sgwm_handle_key_press;
    handlers[KeyRelease] = sgwm_handle_key_release;
    handlers[ButtonPress] = sgwm_handle_button_press;
    handlers[ButtonRelease] = sgwm_handle_button_release;
    handlers[MotionNotify] = sgwm_handle_motion_notify;
    handlers[EnterNotify] = sgwm_handle_enter_notify;
    handlers[LeaveNotify] = sgwm_handle_leave_notify;
    handlers[Expose] = sgwm_handle_expose;
    handlers[VisibilityNotify] = sgwm_handle_visibility_notify;
    handlers[CreateNotify] = sgwm_handle_create_notify;
    handlers[DestroyNotify] = sgwm_handle_destroy_notify;
    handlers[UnmapNotify] = sgwm_handle_unmap_notify;
    handlers[MapNotify] = sgwm_handle_map_notify;
    handlers[MapRequest] = sgwm_handle_map_request;
    handlers[ConfigureRequest] = sgwm_handle_configure_request;
    handlers[PropertyNotify] = sgwm_handle_property_notify;
    handlers[ColormapNotify] = sgwm_handle_colormap_notify;
    handlers[ClientMessage] = sgwm_handle_client_message;
    handlers[SelectionClear] = sgwm_handle_selection_clear;
    handlers[SelectionRequest] = sgwm_handle_selection_request;
}

static void sgwmwidget_map(GtkWidget *widget)
{
    SGWMWidget *sgwm;

    sgwm = SGWMWIDGET(widget);
    GTK_WIDGET_SET_FLAGS(widget,GTK_MAPPED);
}

static void sgwmwidget_show(GtkWidget *widget)
{
    SGWMWidget *sgwm;

    GTK_WIDGET_SET_FLAGS(widget,GTK_VISIBLE);
    if (!GTK_WIDGET_REALIZED(widget)) {
	gtk_widget_realize(widget);
    }
    gtk_widget_map(widget);
    sgwm = SGWMWIDGET(widget);
    gtk_widget_show(sgwm->top);
}

static void sgwmwidget_init_properties(SGWMWidget *sgwm)
{
    GdkWindowAttr attr;
    Atom n = None;
    Window xleader;
    int i;
    guint32 val;
    char *names[SG_N_WORKSPACE];
    Atom win_protocols[SG_N_WIN_PROTOCOLS];
    XTextProperty text;

    attr.width = 0;
    attr.height = 0;
    attr.window_type = GDK_WINDOW_TOPLEVEL;
    attr.wclass = GDK_INPUT_OUTPUT;
    attr.event_mask = 0;
    sgwm->leader = gdk_window_new(NULL,&attr,0);
    xleader = GDK_WINDOW_XWINDOW(sgwm->leader);
    gdk_window_add_filter(sgwm->leader,sgwmwidget_filter,(gpointer)sgwm);
    XChangeProperty(GDK_DISPLAY(),xleader,
      xa_win_supporting_wm_check,
      XA_CARDINAL, 32,
      PropModeReplace, (unsigned char *)&xleader,1);
    XChangeProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),xa_win_supporting_wm_check,
      XA_CARDINAL, 32,
      PropModeReplace, (unsigned char *)&xleader,1);
    XChangeProperty(GDK_DISPLAY(),xleader,xa_sg_wm, XA_ATOM, 32,
      PropModeReplace, (unsigned char *)&(n),1);
    i = 0;
    win_protocols[i++] = xa_win_state;
    win_protocols[i++] = xa_win_hints;
    win_protocols[i++] = xa_win_workspace;
    win_protocols[i++] = xa_win_workspace_count;
    win_protocols[i++] = xa_win_workspace_names;
    win_protocols[i++] = xa_win_client_list;
    win_protocols[i++] = xa_sgws_swap;
    win_protocols[i++] = xa_sgws_enable;
    win_protocols[i++] = xa_sgws_activate;
    win_protocols[i++] = xa_sgws_resize;
    win_protocols[i++] = xa_sgws_mapws;
    XChangeProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),xa_win_protocols,
      XA_ATOM,32,PropModeReplace,(unsigned char *)win_protocols,i);
    val = SG_N_WORKSPACE;
    XChangeProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),xa_win_workspace_count,
      XA_CARDINAL,32,PropModeReplace,(unsigned char *)&val,1);
    sgwindow_set_workspace_property(NULL,0);
    names[0] = "Main";
#if (SG_N_WORKSPACE == 2)
    names[1] = "Sub";
#elif (SG_N_WORKSPACE >= 3)
    for (i = 2;i < SG_N_WORKSPACE;i++) {
	char buf[256];

	sprintf(buf,"Sub%d",i-1);
	names[i] = g_strdup(buf);
    }
#endif
    if (XStringListToTextProperty(names,SG_N_WORKSPACE,&text)) {
	XSetTextProperty(GDK_DISPLAY(),GDK_ROOT_WINDOW(),
	  &text,xa_win_workspace_names);
	XFree(text.value);
    }
#if (SG_N_WORKSPACE >= 3)
    for (i = 2;i < SG_N_WORKSPACE;i++) {
	g_free(names[i]);
    }
#endif
    sgwmwidget_set_client_list_property(sgwm);
}

static void sgwmwidget_init_selections(SGWMWidget *sgwm)
{
    Window owner;
    XClientMessageEvent ev;
    Window xleader;

    xleader = GDK_WINDOW_XWINDOW(sgwm->leader);
    do {
	owner = XGetSelectionOwner(GDK_DISPLAY(),xa_wm_s);
	if (owner != None) {
	    g_error("Other Window Manager is already running\n");
	    exit(2);
	}
	XSetSelectionOwner(GDK_DISPLAY(),xa_wm_s,xleader,CurrentTime);
	owner = XGetSelectionOwner(GDK_DISPLAY(),xa_wm_s);
    } while (owner != xleader);
    ev.type = ClientMessage;
    ev.message_type = xa_manager;
    ev.format = 32;
    ev.window = GDK_ROOT_WINDOW();
    ev.data.l[0] = CurrentTime;
    ev.data.l[1] = xa_wm_s;
    ev.data.l[2] = xleader;
    ev.data.l[3] = 0;
    ev.data.l[4] = 0;
    XSendEvent(GDK_DISPLAY(),GDK_ROOT_WINDOW(),False,StructureNotifyMask,
      (XEvent *)&ev);
}

static void sgwmwidget_realize(GtkWidget *widget)
{
    SGWMWidget *sgwm;
    gint width;
    gint height;
    GtkWidget *label;
    GtkWidget *frame;

    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);
    sgwm = SGWMWIDGET(widget);
    widget->window = gdk_window_foreign_new(GDK_ROOT_WINDOW());
    gdk_window_set_user_data(widget->window,widget);
    gdk_window_add_filter(widget->window,sgwmwidget_filter,(gpointer)sgwm);
    gdk_add_client_message_filter(xa_wm_change_state,sgwm_client_message_filter,
      (gpointer)sgwm);
    gdk_error_trap_push();
    XSelectInput(GDK_DISPLAY(), GDK_ROOT_WINDOW(),
	SubstructureRedirectMask | KeyPressMask 
	| KeyReleaseMask |SubstructureNotifyMask);
    XDefineCursor(GDK_DISPLAY(),GDK_ROOT_WINDOW(),defaultCursor);
    XSync(GDK_DISPLAY(), 0);
    if (gdk_error_trap_pop()) {
	fprintf(stderr,"Other Window Manager is already running\n");
	exit(2);
    }
    sgwmwidget_init_handlers();
    gdk_window_get_size(widget->window,&width,&height);
    widget->allocation.x = 0;
    widget->allocation.y = 0;
    widget->allocation.width = width;
    widget->allocation.height = height;
    widget->requisition.width = width;
    widget->requisition.height = height;
    sgwm->top = sgwmtop_new(sgwm);
    gtk_widget_size_allocate(sgwm->top,&(widget->allocation));
    gtk_widget_realize(sgwm->top);
    sgwmwidget_init_properties(sgwm);
    sgwmwidget_init_selections(sgwm);
    sgwm->prompt = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_window_set_position(GTK_WINDOW(sgwm->prompt),GTK_WIN_POS_CENTER);
    frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(sgwm->prompt),frame);
#ifdef __SIKIGAMI_PDA__
    label = gtk_label_new(_("Please tap the area you want to map the task."));
#else
    label = gtk_label_new(_("Please tap the area you want to map the task. Or enter key 0(Main area),1(Sub area)."));
#endif
    gtk_container_add(GTK_CONTAINER(frame),label);
}

typedef void (*SGSignal_NONE__INT_INT_INT)(GtkObject *object,
  guint arg1, guint arg2, guint arg3, gpointer user_data);

static void sg_marshal_NONE__INT_INT_INT(GtkObject * object,
                                  GtkSignalFunc func,
                                  gpointer func_data,
                                  GtkArg * args)
{
    SGSignal_NONE__INT_INT_INT rfunc;

    rfunc = (SGSignal_NONE__INT_INT_INT)func;
    (*rfunc)(object,
        GTK_VALUE_INT (args[0]),
        GTK_VALUE_INT (args[1]),
        GTK_VALUE_INT (args[2]),
        func_data);
}

static void sgwmwidget_class_init(SGWMWidgetClass *klass)
{
    GtkWidgetClass *widget_class;
    GtkObjectClass *object_class;

    widget_class = (GtkWidgetClass *)klass;
    widget_class->map = sgwmwidget_map;
    widget_class->realize = sgwmwidget_realize;
    widget_class->show = sgwmwidget_show;
    queryWSCursor = XCreateFontCursor(GDK_DISPLAY(),XC_draped_box);
    defaultCursor = XCreateFontCursor(GDK_DISPLAY(),XC_left_ptr);

    object_class = (GtkObjectClass *)klass;
    sgwm_signals[SWAP_TASK] = gtk_signal_new("swap_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,swap_task),
				       gtk_marshal_NONE__INT_INT,
				       GTK_TYPE_NONE, 2,
				       GTK_TYPE_INT,
				       GTK_TYPE_INT);
    sgwm_signals[ENABLE_WORKSPACE] = gtk_signal_new("enable_workspace",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         enable_workspace),
				       gtk_marshal_NONE__INT_INT,
				       GTK_TYPE_NONE, 2,
				       GTK_TYPE_INT,
				       GTK_TYPE_INT);
    sgwm_signals[ACTIVATE_WORKSPACE] = gtk_signal_new("activate_workspace",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         activate_workspace),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[RESIZE_WORKSPACE] = gtk_signal_new("resize_workspace",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         resize_workspace),
				       sg_marshal_NONE__INT_INT_INT,
				       GTK_TYPE_NONE, 3,
				       GTK_TYPE_INT,
				       GTK_TYPE_INT,
				       GTK_TYPE_INT);
    sgwm_signals[CALL_DESKTOP] = gtk_signal_new("call_desktop",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         call_desktop),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[MAP_TASK_TO_WORKSPACE] = gtk_signal_new("map_task_to_workspace",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         map_task_to_workspace),
				       gtk_marshal_NONE__INT_INT,
				       GTK_TYPE_NONE, 2,
				       GTK_TYPE_INT,
				       GTK_TYPE_INT);
    sgwm_signals[MAP_TASK] = gtk_signal_new("map_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         map_task),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[ROLL_UP_TASK] = gtk_signal_new("roll_up_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         roll_up_task),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[ROLL_DOWN_TASK] = gtk_signal_new("roll_down_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         roll_down_task),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[MOVE_NEXT_TASK] = gtk_signal_new("move_next_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         move_next_task),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    sgwm_signals[MOVE_PREV_TASK] = gtk_signal_new("move_prev_task",
				       GTK_RUN_FIRST | GTK_RUN_ACTION,
				       object_class->type,
				       GTK_SIGNAL_OFFSET(SGWMWidgetClass,
				         move_prev_task),
				       gtk_marshal_NONE__INT,
				       GTK_TYPE_NONE, 1,
				       GTK_TYPE_INT);
    gtk_object_class_add_signals(object_class, sgwm_signals,LAST_SIGNAL);
    gtk_object_class_add_signals(object_class, sgwm_signals,LAST_SIGNAL);

    klass->enable_workspace = sgwmwidget_enable_workspace;
    klass->activate_workspace = sgwmwidget_activate_workspace;
    klass->resize_workspace = sgwmwidget_resize_workspace;
    klass->swap_task = sgwmwidget_swap_task;
    klass->roll_up_task = sgwmwidget_roll_up_task;
    klass->roll_down_task = sgwmwidget_roll_down_task;
    klass->move_next_task = sgwmwidget_move_next_task;
    klass->move_prev_task = sgwmwidget_move_prev_task;
    klass->map_task = sgwmwidget_map_task;
    klass->map_task_to_workspace = sgwmwidget_map_task_to_workspace;
    klass->call_desktop = sgwmwidget_call_desktop;
}

static void sgwmwidget_init(SGWMWidget *sgwm)
{
    sgwm->top = NULL;
    sgwm->client_list = NULL;
    sgwm->client_list_changed = FALSE;
    sgwm->prompt = NULL;
    sgwm->desktop = NULL;
}

GtkWidget *sgwmwidget_new(void)
{
    return GTK_WIDGET(gtk_type_new(sgwmwidget_get_type()));
}

void sgwmwidget_do_activate_workspace(SGWMWidget *sgwm, gint no,
  gboolean warp, gboolean focus)
{
    sgwmtop_activate_workspace(SGWMTOP(sgwm->top),no,warp,focus);
}

void sgwmwidget_exit(int status)
{
    XSelectInput(GDK_DISPLAY(), GDK_ROOT_WINDOW(),0);
    XSync(GDK_DISPLAY(), 0);
    exit(status);
}

static void sgwmwidget_set_default_binding(SGWMWidget *sgwm,
   GtkBindingSet *binding_set)
{
    gtk_binding_entry_add_signal(binding_set,
				GDK_F5, GDK_MOD1_MASK,
				"call_desktop",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_F3, GDK_MOD1_MASK,
				"swap_task",
				2,
				GTK_TYPE_INT, 0,
				GTK_TYPE_INT, 1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_F4, GDK_MOD1_MASK,
				"activate_workspace",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Up, GDK_MOD1_MASK,
				"roll_up_task",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Down, GDK_MOD1_MASK,
				"roll_down_task",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Right, GDK_MOD1_MASK,
				"move_next_task",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Left, GDK_MOD1_MASK,
				"move_prev_task",
				1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Delete, GDK_MOD1_MASK,
				"enable_workspace",
				2,
				GTK_TYPE_INT, 1,
				GTK_TYPE_INT, -1);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Page_Up, GDK_MOD1_MASK,
				"resize_workspace",
				3,
				GTK_TYPE_INT, 0,
				GTK_TYPE_INT, 1,
				GTK_TYPE_INT, 10);
    gtk_binding_entry_add_signal(binding_set,
				GDK_Page_Down, GDK_MOD1_MASK,
				"resize_workspace",
				3,
				GTK_TYPE_INT, 0,
				GTK_TYPE_INT, 1,
				GTK_TYPE_INT, -10);
}

static void sgwmwidget_grab_keys(SGWMWidget *sgwm, 
    GtkBindingSet *binding_set)
{
    /* Notice: Using non-public interface of GtkBindingSet */

    GtkBindingEntry *ent;

    for (ent = binding_set->entries;ent != NULL;ent = ent->set_next) {
	XGrabKey(GDK_DISPLAY(),XKeysymToKeycode(GDK_DISPLAY(), ent->keyval),
	  ent->modifiers,GDK_ROOT_WINDOW(),
	  False,GrabModeAsync,GrabModeAsync);
    }
    if (middleButtonKeycode != 0) {
	XGrabKey(GDK_DISPLAY(),middleButtonKeycode,
	  AnyModifier,GDK_ROOT_WINDOW(),
	  False,GrabModeAsync,GrabModeAsync);
    }
    if (rightButtonKeycode != 0) {
	XGrabKey(GDK_DISPLAY(),rightButtonKeycode,
	  AnyModifier,GDK_ROOT_WINDOW(),
	  False,GrabModeAsync,GrabModeAsync);
    }
    if (middleButtonKeycode != 0 || rightButtonKeycode != 0) {
	nbuttons = XGetPointerMapping(GDK_DISPLAY(),NULL,0);
    }
}

void sgwmwidget_start(SGWMWidget *sgwm)
{
    Window *children;
    Window root,parent;
    Window tw;
    guint nc;
    int i;
    int ws;
    GtkBindingSet *binding_set;

    tw = GDK_WINDOW_XWINDOW(GTK_WIDGET(sgwm->top)->window);
    XQueryTree(GDK_DISPLAY(),GDK_ROOT_WINDOW(),&root,&parent,&children,&nc);
    ws = 0;
    for (i = 0;i < nc;i++) {
	if (children[i] != tw) {
	    XWindowAttributes attr;
	    guint32 data[1];
	    Window *tmpchild;
	    guint tmpnc;

	    gdk_error_trap_push();
	    XGetWindowAttributes(GDK_DISPLAY(),children[i],&attr);
	    if (XQueryTree(GDK_DISPLAY(),children[i],&root,&parent,
	      &tmpchild,&tmpnc) != 0) {
		if (parent == GDK_ROOT_WINDOW()
		  && attr.map_state != IsUnmapped
		  && !attr.override_redirect) {
		    XUnmapWindow(GDK_DISPLAY(),children[i]);
		    if (!sgwmwidget_is_barwindow(children[i])) {
			data[0] = ws++ % SG_N_WORKSPACE;
			XChangeProperty(GDK_DISPLAY(),children[i],
			  xa_win_workspace,XA_CARDINAL,32,PropModeReplace,
			    (unsigned char *)data,1);
			sgwmwidget_do_map_task(sgwm,children[i]);
		    }
		}
		if (tmpchild != NULL) XFree(tmpchild);
	    }
	    gdk_flush();
	    if (gdk_error_trap_pop()) {
		;
	    }
	}
    }
    if (children != NULL) XFree(children);
    /* Notice: Using Predefined name binding set. */
    binding_set = gtk_binding_set_find(SGWM_GRABBED_BINDING);
    if (binding_set == NULL) {
	binding_set = gtk_binding_set_by_class(gtk_type_class(GTK_TYPE_SGWMWIDGET));
	sgwmwidget_set_default_binding(sgwm,binding_set);
    }
    sgwmwidget_grab_keys(sgwm,binding_set);
}

